@@ -1,4 +1,5 @@
 ((nil . ((indent-tabs-mode . nil)
+         (fill-column 100)
          (qpdf-cc-style
           .
           ("qpdf"
@@ -36,4 +37,9 @@
                     )
             ))
  )
+ (gfm-mode . ((eval . (progn
+                        (setq fill-column 100)
+                        )
+                    ))
+ )
 )
