@@ -21,12 +21,9 @@ void
 BitWriter::writeBitsSigned(long long val, size_t bits)
 {
     unsigned long long uval = 0;
-    if (val < 0)
-    {
+    if (val < 0) {
         uval = (1ULL << bits) + static_cast<unsigned long long>(val);
-    }
-    else
-    {
+    } else {
         uval = static_cast<unsigned long long>(val);
     }
     writeBits(uval, bits);
@@ -41,8 +38,7 @@ BitWriter::writeBitsInt(int val, size_t bits)
 void
 BitWriter::flush()
 {
-    if (bit_offset < 7)
-    {
+    if (bit_offset < 7) {
         size_t bits_to_write = bit_offset + 1;
         write_bits(this->ch, this->bit_offset, 0, bits_to_write, this->pl);
     }
