@@ -510,6 +510,20 @@ QPDFJob::Config::removeAttachment(std::string const& parameter)
     return this;
 }
 
+QPDFJob::Config*
+QPDFJob::Config::removeInfo()
+{
+    o.m->remove_info = true;
+    return this;
+}
+
+QPDFJob::Config*
+QPDFJob::Config::removeMetadata()
+{
+    o.m->remove_metadata = true;
+    return this;
+}
+
 QPDFJob::Config*
 QPDFJob::Config::removePageLabels()
 {
