@@ -12,11 +12,15 @@ import sys
 
 sys.path.append(os.path.abspath("./_ext"))
 
-project = 'QPDF'
-copyright = '2005-2023, Jay Berkenbilt'
+project = 'qpdf'
+copyright = '2005-2021 Jay Berkenbilt, 2022-2025 Jay Berkenbilt and Manfred Holger'
 author = 'Jay Berkenbilt'
-# make_dist and the CI build lexically find the release version from this file.
-release = '11.6.4'
+here = os.path.dirname(os.path.realpath(__file__))
+with open(f'{here}/../CMakeLists.txt') as f:
+    for line in f.readlines():
+        if line.strip().startswith('VERSION '):
+            release = line.replace('VERSION', '').strip()
+            break
 version = release
 extensions = [
     'sphinx_rtd_theme',
